import argparse
import json
import os
import re
import sys
from configparser import RawConfigParser
from dataclasses import dataclass, field, fields
from urllib.parse import ParseResult, urlparse

import myloginpath
from rich import box
from rich.console import Console
from rich.table import Table
from rich.theme import Theme

from dolphie.DataTypes import Panels
from dolphie.Modules.Queries import MySQLQueries


@dataclass
class CredentialProfile:
    name: str
    host: str = None
    port: int = None
    user: str = None
    password: str = None
    socket: str = None
    ssl: dict = field(default_factory=dict)
    ssl_mode: str = None
    ssl_ca: str = None
    ssl_cert: str = None
    ssl_key: str = None


@dataclass
class HostGroupMember:
    tab_title: str
    host: str
    port: int = None
    credential_profile: CredentialProfile = None


@dataclass
class Config:
    app_version: str
    tab_setup: bool = False
    credential_profile: str = None
    user: str = None
    password: str = None
    host: str = "localhost"
    port: int = 3306
    socket: str = None
    ssl: dict = field(default_factory=dict)
    ssl_mode: str = None
    ssl_ca: str = None
    ssl_cert: str = None
    ssl_key: str = None
    config_file: list[str] = field(
        default_factory=lambda: [
            "/etc/dolphie.cnf",
            "/etc/dolphie/dolphie.cnf",
            f"{os.path.expanduser('~')}/.dolphie.cnf",
        ]
    )
    mycnf_file: str = field(default_factory=lambda: f"{os.path.expanduser('~')}/.my.cnf")
    login_path: str = "client"
    host_cache_file: str = field(default_factory=lambda: f"{os.path.expanduser('~')}/dolphie_host_cache")
    tab_setup_file: str = field(default_factory=lambda: f"{os.path.expanduser('~')}/dolphie_hosts")
    refresh_interval: int = 1
    heartbeat_table: str = None
    credential_profiles: dict[str, CredentialProfile] = field(default_factory=dict)
    tab_setup_available_hosts: list[str] = field(default_factory=list)
    startup_panels: list[str] = field(default_factory=lambda: ["dashboard", "processlist"])
    graph_marker: str = "braille"
    pypi_repository: str = "https://pypi.org/pypi/dolphie/json"
    hostgroup: str = None
    hostgroup_hosts: dict[str, list[HostGroupMember]] = field(default_factory=dict)
    show_trxs_only: bool = False
    show_additional_query_columns: bool = False
    record_for_replay: bool = False
    daemon_mode: bool = False
    daemon_mode_panels: list[str] = field(default_factory=lambda: ["processlist", "metadata_locks"])
    daemon_mode_log_file: str = field(default_factory=lambda: f"{os.path.expanduser('~')}/dolphie_daemon.log")
    replay_file: str = None
    replay_dir: str = None
    replay_retention_hours: int = 48
    exclude_notify_global_vars: str = None


class ArgumentParser:
    def __init__(self, app_version: str):
        self.config_object_options = {}

        for variable in fields(Config):
            # Exclude these options since we handle them differently
            if variable.name not in [
                "app_version",
                "tab_setup_available_hosts",
                "ssl",
                "hostgroup_hosts",
                "credential_profiles",
            ]:
                self.config_object_options[variable.name] = variable.type

        self.formatted_options = "\n\t".join(
            [
                (
                    f"(comma-separated str) {option}"
                    if option in ("daemon_mode_panels", "startup_panels", "exclude_notify_global_vars")
                    else f"({data_type.__name__}) {option}" if hasattr(data_type, "__name__") else f"(str) {option} []"
                )
                for option, data_type in self.config_object_options.items()
                if option != "config_file"
            ]
        )

        epilog = f"""
Order of precedence for methods that pass options to Dolphie:
\t1. Command-line
\t2. Credential profile (set by --cred-profile)
\t3. Environment variables
\t4. Dolphie's config (set by --config-file)
\t5. ~/.mylogin.cnf (mysql_config_editor)
\t6. ~/.my.cnf (set by --mycnf-file)

Credential profiles can be defined in Dolphie's config file as a way to store credentials for easy access.
A profile can be created by adding a section in the config file with the format: [credential_profile_<name>]
When using a credential profile, do not include the prefix 'credential_profile' (i.e. -C production)
The following options are supported in credential profiles:
\thost
\tport (default is 3306)
\tuser
\tpassword
\tsocket
\tssl_mode REQUIRED/VERIFY_CA/VERIFY_IDENTITY
\tssl_ca
\tssl_cert
\tssl_key
\tmycnf_file
\tlogin_path

MySQL my.cnf file supports these options under [client] section:
\thost
\tuser
\tpassword
\tport
\tsocket
\tssl_mode REQUIRED/VERIFY_CA/VERIFY_IDENTITY
\tssl_ca
\tssl_cert
\tssl_key

Login path file supports these options:
\thost
\tuser
\tpassword
\tport
\tsocket

Environment variables support these options:
\tDOLPHIE_USER
\tDOLPHIE_PASSWORD
\tDOLPHIE_HOST
\tDOLPHIE_PORT
\tDOLPHIE_SOCKET
\tDOLPHIE_SSL_MODE REQUIRED/VERIFY_CA/VERIFY_IDENTITY
\tDOLPHIE_SSL_CA
\tDOLPHIE_SSL_CERT
\tDOLPHIE_SSL_KEY

Dolphie's config supports these options under [dolphie] section:
\t{self.formatted_options}
"""
        self.parser = argparse.ArgumentParser(
            conflict_handler="resolve",
            description="Dolphie, an intuitive feature-rich top tool for monitoring MySQL in real time",
            epilog=epilog,
            formatter_class=argparse.RawTextHelpFormatter,
        )
        self.config = Config(app_version)
        self.panels = Panels()

        self.console = Console(style="#e9e9e9", highlight=False)
        self.console.push_theme(
            Theme(
                {
                    "red2": "b #fb9a9a",
                }
            )
        )

        self._add_options()
        self._parse()

    def _add_options(self):
        self.parser.add_argument(
            "uri",
            metavar="uri",
            type=str,
            nargs="?",
            help=(
                "Use a URI string for credentials (mysql/proxysql) - format: mysql://user:password@host:port "
                f"(port is optional with default {self.config.port}, or 6032 for ProxySQL)"
            ),
        )

        self.parser.add_argument(
            "--tab-setup",
            dest="tab_setup",
            action="store_true",
            help=(
                "Start Dolphie by showing the Tab Setup modal instead of automatically connecting "
                "with the specified options"
            ),
        )
        self.parser.add_argument(
            "-C",
            "--cred-profile",
            dest="credential_profile",
            type=str,
            help="Credential profile to use. See below for more information",
            metavar="",
        )
        self.parser.add_argument("-u", "--user", dest="user", type=str, help="Username", metavar="")
        self.parser.add_argument("-p", "--password", dest="password", type=str, help="Password", metavar="")
        self.parser.add_argument(
            "-h",
            "--host",
            dest="host",
            type=str,
            help="Hostname/IP address",
            metavar="",
        )
        self.parser.add_argument(
            "-P",
            "--port",
            dest="port",
            type=int,
            help="Port (socket has precedence)",
            metavar="",
        )
        self.parser.add_argument(
            "-S",
            "--socket",
            dest="socket",
            type=str,
            help="Socket file",
            metavar="",
        )
        self.parser.add_argument(
            "-c",
            "--config-file",
            dest="config_file",
            type=str,
            help=(
                f"Dolphie's config file to use. Options are read from these files in the given order: "
                f"{self.config.config_file}"
            ),
            metavar="",
        )
        self.parser.add_argument(
            "-m",
            "--mycnf-file",
            dest="mycnf_file",
            type=str,
            help=(
                "MySQL config file path to use. This should use [client] section "
                f"[default: {self.config.mycnf_file}]"
            ),
            metavar="",
        )
        self.parser.add_argument(
            "-l",
            "--login-path",
            dest="login_path",
            type=str,
            help=(
                "Specify login path to use with mysql_config_editor's file ~/.mylogin.cnf for encrypted login "
                f"credentials [default: {self.config.login_path}]"
            ),
            metavar="",
        )
        self.parser.add_argument(
            "-r",
            "--refresh-interval",
            dest="refresh_interval",
            type=int,
            help=(
                "The time, in seconds, between each data collection and processing cycle "
                f"[default: {self.config.refresh_interval}]"
            ),
            metavar="",
        )
        self.parser.add_argument(
            "--host-cache-file",
            dest="host_cache_file",
            type=str,
            help=(
                "Resolve IPs to hostnames when your DNS is unable to. Each IP/hostname pair should be on its own line "
                f"using format ip=hostname [default: {self.config.host_cache_file}]"
            ),
            metavar="",
        )
        self.parser.add_argument(
            "--tab-setup-file",
            dest="tab_setup_file",
            type=str,
            help=(
                "Specify location of file that stores the available hosts to use in Tab Setup modal [default: "
                f"{self.config.tab_setup_file}]"
            ),
            metavar="",
        )
        self.parser.add_argument(
            "--heartbeat-table",
            dest="heartbeat_table",
            type=str,
            help=(
                "(MySQL only) If your hosts use pt-heartbeat, specify table in format db.table to use the timestamp it "
                "has for replication lag instead of Seconds_Behind_Master from SHOW REPLICA STATUS"
            ),
            metavar="",
        )
        self.parser.add_argument(
            "--ssl-mode",
            dest="ssl_mode",
            type=str,
            help=(
                "Desired security state of the connection to the host. Supports: "
                "REQUIRED/VERIFY_CA/VERIFY_IDENTITY [default: OFF]"
            ),
            metavar="",
        )
        self.parser.add_argument(
            "--ssl-ca",
            dest="ssl_ca",
            type=str,
            help="Path to the file that contains a CA (certificate authority)",
            metavar="",
        )
        self.parser.add_argument(
            "--ssl-cert",
            dest="ssl_cert",
            type=str,
            help="Path to the file that contains a certificate",
            metavar="",
        )
        self.parser.add_argument(
            "--ssl-key",
            dest="ssl_key",
            type=str,
            help="Path to the file that contains a private key for the certificate",
            metavar="",
        )
        self.parser.add_argument(
            "--panels",
            dest="startup_panels",
            type=str,
            help=(
                "What panels to display on startup separated by a comma. Supports: "
                f"{self.panels.all()}, [default: {self.config.startup_panels}]"
            ),
            metavar="",
        )
        self.parser.add_argument(
            "--graph-marker",
            dest="graph_marker",
            type=str,
            help=(
                "What marker to use for graphs (available options: https://tinyurl.com/dolphie-markers) [default: "
                f"{self.config.graph_marker}]"
            ),
            metavar="",
        )
        self.parser.add_argument(
            "--pypi-repo",
            dest="pypi_repository",
            type=str,
            help=(
                "What PyPi repository to use when checking for a new version "
                f"default: [{self.config.pypi_repository}]"
            ),
            metavar="",
        )
        self.parser.add_argument(
            "-H",
            "--hostgroup",
            dest="hostgroup",
            type=str,
            help=(
                "This is used for creating tabs and connecting to them for hosts you specify in "
                "Dolphie's config file under a hostgroup section. As an example, you'll have a section "
                "called [cluster1] then below it you will list each host on a new line in the format "
                "key=host (keys have no meaning). Hosts support optional port (default is whatever port parameter is) "
                "in the format host:port. You can also name the tabs by suffixing "
                "~tab_name to the host (i.e. 1=host~tab_name)"
            ),
            metavar="",
        )
        self.parser.add_argument(
            "-R",
            "--record",
            dest="record_for_replay",
            action="store_true",
            help=(
                "Enables recording of Dolphie's data to a replay file. "
                "Note: This can use significant disk space. Monitor accordingly!"
            ),
        )
        self.parser.add_argument(
            "-D",
            "--daemon",
            dest="daemon_mode",
            action="store_true",
            help=(
                "Starts Dolphie in daemon mode. This will not show the TUI and is designed be put into the "
                "background with whatever solution you decide to use. Automatically enables --record. "
                "This mode is solely used for recording data to a replay file"
            ),
        )
        self.parser.add_argument(
            "--daemon-log-file",
            dest="daemon_mode_log_file",
            type=str,
            help="Full path of the log file for daemon mode",
            metavar="",
        )
        self.parser.add_argument(
            "--daemon-panels",
            dest="daemon_mode_panels",
            type=str,
            help=(
                "Which panels to run queries for in daemon mode separated by a comma. Reducing panels can lower "
                f"query load and replay file size. Dashboard/Replication panels cannot be turned off. Supports: "
                f"{self.panels.get_all_daemon_panel_names()}, [default: {self.config.daemon_mode_panels}]"
            ),
            metavar="",
        )
        self.parser.add_argument(
            "--replay-file",
            dest="replay_file",
            type=str,
            help="Specify the full path of the replay file to load and enable replay mode",
            metavar="",
        )
        self.parser.add_argument(
            "--replay-dir",
            dest="replay_dir",
            type=str,
            help="Directory to store replay data files",
            metavar="",
        )
        self.parser.add_argument(
            "--replay-retention-hours",
            dest="replay_retention_hours",
            type=int,
            help=(
                f"Number of hours to keep replay data. Data will be purged every hour "
                f"[default: {self.config.replay_retention_hours}]"
            ),
            metavar="",
        )
        self.parser.add_argument(
            "--exclude-notify-vars",
            dest="exclude_notify_global_vars",
            type=str,
            help=(
                "Dolphie will let you know when a global variable has been changed. If you have variables that change "
                "frequently and you don't want to see them, you can specify which ones with this option separated by "
                "a comma (i.e. --exclude-notify-vars=variable1,variable2)"
            ),
            metavar="",
        )
        self.parser.add_argument(
            "--show-trxs-only",
            dest="show_trxs_only",
            action="store_true",
            help="(MySQL only) Start with only showing threads that have an active transaction",
        )
        self.parser.add_argument(
            "--additional-columns",
            dest="show_additional_query_columns",
            action="store_true",
            help="Start with additional columns in Processlist panel",
        )
        self.parser.add_argument(
            "--debug-options",
            dest="debug_options",
            action="store_true",
            help=(
                "Display options that are set and what they're set by (command-line, dolphie config, etc) then exit. "
                "WARNING: This will show passwords and other sensitive information in plain text"
            ),
        )
        self.parser.add_argument(
            "-V",
            "--version",
            action="version",
            version=self.config.app_version,
            help="Display version and exit",
        )

    def set_config_value(self, source, option, value):
        setattr(self.config, option, value)
        self.add_to_debug_options(source, option, value)

    def add_to_debug_options(self, source, option, value):
        if self.debug_options:
            self.debug_options_table.add_row(source, option, str(value))

    def _parse(self):
        login_options = ["user", "password", "host", "port", "socket", "ssl_mode", "ssl_ca", "ssl_cert", "ssl_key"]

        options = vars(self.parser.parse_args())  # Convert object to dictionary

        dolphie_config_login_options_used = {}
        hostgroups = {}

        self.debug_options = False
        if options["debug_options"]:
            self.debug_options = True

            self.debug_options_table = Table(box=box.SIMPLE_HEAVY, header_style="b", style="#333f62")
            self.debug_options_table.add_column("Source")
            self.debug_options_table.add_column("Option", style="#91abec")
            self.debug_options_table.add_column("Value", style="#bbc8e8")

        for option in self.config_object_options:
            if self.debug_options:
                self.debug_options_table.add_row("default", option, str(getattr(self.config, option)))

                # If last option, add a separator
                if option == list(self.config_object_options.keys())[-1]:
                    self.debug_options_table.add_row("", "", "")

        if options["config_file"]:
            self.config.config_file = [options["config_file"]]

        # Loop through config files to find the supplied options
        for config_file in self.config.config_file:
            if os.path.isfile(config_file):
                cfg = RawConfigParser()
                cfg.read(config_file)

                # Loop through all of available options
                for option, data_type in self.config_object_options.items():
                    # If the option is in the config file
                    if cfg.has_option("dolphie", option):
                        # Check if the value is of the correct data type
                        value = self.verify_config_value(option, cfg.get("dolphie", option), data_type)

                        # If the option is not a login option, save it to the config object
                        if option not in login_options and value:
                            self.set_config_value("dolphie config", option, value)
                        else:
                            # Save the login option to be used later
                            dolphie_config_login_options_used[option] = value

                # First, loop sections for credential profiles so hostgroups can reference them
                # when the profiles are not before hostgroup declarations in Dolphie's config file
                for section in cfg.sections():
                    if section.startswith("credential_profile"):
                        self.parse_credential_profile(cfg, section)

                # Then, loop sections for hostgroups
                for section in cfg.sections():
                    if section == "dolphie" or section.startswith("credential_profile"):
                        continue

                    # Treat anything else as a hostgroup
                    hosts = self.parse_hostgroup(cfg, section, config_file)
                    if hosts:
                        hostgroups[section] = hosts

        # Save the hostgroups found to the config object
        self.config.hostgroup_hosts = hostgroups

        # We need to loop through all options and set non-login options so we can use them for the logic below
        for option in self.config_object_options:
            if option not in login_options and options[option]:
                self.set_config_value("command-line", option, options[option])

        if self.config.credential_profile and self.config.credential_profile not in self.config.credential_profiles:
            self.exit(
                f"Credential profile [red2]{self.config.credential_profile}[/red2] does not exist in "
                "Dolphie's config file"
            )

        # Use MySQL's my.cnf file for login options if specified
        if os.path.isfile(self.config.mycnf_file):
            cfg = RawConfigParser()
            cfg.read(self.config.mycnf_file)

            for option in login_options:
                if cfg.has_option("client", option):
                    self.set_config_value("my.cnf", option, cfg.get("client", option))

        # Use login path for login options if specified
        if self.config.login_path:
            try:
                login_path_data = myloginpath.parse(self.config.login_path)

                for option in login_options:
                    if option in login_path_data:
                        self.set_config_value("login path", option, login_path_data[option])
            except Exception as e:
                # Don't error out for default login path
                if self.config.login_path != "client":
                    self.exit(f"Problem reading login path file: {e}")

        # Loop through all login options and set them in order of precedence
        for option in login_options:
            # Update config object with Dolphie config
            dolphie_value = dolphie_config_login_options_used.get(option)
            if dolphie_value:
                self.set_config_value("dolphie config", option, dolphie_value)

            # Use environment variables if specified
            environment_var = f"DOLPHIE_{option.upper()}"
            env_value = os.environ.get(environment_var)
            if env_value:
                self.set_config_value("env variable", option, env_value)

            # Use credential profile if specified
            if self.config.credential_profile:
                profile = self.config.credential_profiles.get(self.config.credential_profile)
                if profile and hasattr(profile, option) and getattr(profile, option):
                    self.set_config_value(f"cred profile {profile.name}", option, getattr(profile, option))

            # Use command-line arguments if specified
            if options.get(option):
                self.set_config_value("command-line", option, options[option])

        # Create SSL object from config after all login options are set
        ssl_payload = {
            opt.name: getattr(self.config, opt.name)
            for opt in fields(Config)
            if opt.name.startswith("ssl_") and getattr(self.config, opt.name)
        }
        if ssl_payload:
            self.set_config_value("ssl object", "ssl", self.create_ssl_object(ssl_payload))

        # Lastly, parse URI if specified
        if options["uri"]:
            try:
                parsed_result: ParseResult = urlparse(options["uri"])

                if parsed_result.scheme == "mysql":
                    port = parsed_result.port or 3306
                elif parsed_result.scheme == "proxysql":
                    port = parsed_result.port or 6032
                else:
                    self.exit(
                        "Invalid URI scheme: Only 'mysql' or 'proxysql' are supported (see --help for more information)"
                    )

                self.set_config_value("uri", "user", parsed_result.username)
                self.set_config_value("uri", "password", parsed_result.password)
                self.set_config_value("uri", "host", parsed_result.hostname)
                self.set_config_value("uri", "port", port)
            except Exception as e:
                self.exit(f"Invalid URI: {e} (see --help for more information)")

        if self.config.hostgroup:
            # Sanity check for hostgroup
            if self.config.hostgroup not in hostgroups:
                self.exit(f"Hostgroup [red2]{self.config.hostgroup}[/red2] does not exist in Dolphie's config file")

        if self.config.heartbeat_table:
            pattern_match = re.search(r"^(\w+\.\w+)$", self.config.heartbeat_table)
            if pattern_match:
                MySQLQueries.heartbeat_replica_lag = MySQLQueries.heartbeat_replica_lag.replace(
                    "$1", self.config.heartbeat_table
                )
            else:
                self.exit("Your heartbeat table did not conform to the proper format: db.table")

        if self.config.exclude_notify_global_vars:
            self.config.exclude_notify_global_vars = self.config.exclude_notify_global_vars.split(",")

        # Validate panels
        try:
            self.config.startup_panels = self.panels.validate_panels(self.config.startup_panels, self.panels.all())
            self.config.daemon_mode_panels = self.panels.validate_panels(
                self.config.daemon_mode_panels, self.panels.get_all_daemon_panel_names()
            )
        except ValueError as e:
            self.exit(str(e))

        if os.path.exists(self.config.tab_setup_file):
            with open(self.config.tab_setup_file) as file:
                self.config.tab_setup_available_hosts = [line.strip() for line in file]

        if self.debug_options:
            self.console.print(self.debug_options_table)
            self.console.print("[#969aad]Note: Options are set by their source in the order they appear")
            sys.exit()

        # Verify parameters for replay & daemon mode
        if self.config.daemon_mode:
            self.config.record_for_replay = True
            if not self.config.replay_dir:
                self.exit("Daemon mode ([red2]--daemon[/red2]) requires [red2]--replay-dir[/red2] to be specified")

        if self.config.replay_file and not os.path.isfile(self.config.replay_file):
            self.exit(f"Replay file [red2]{self.config.replay_file}[/red2] does not exist")

        if self.config.record_for_replay and not self.config.replay_dir:
            self.exit("[red2]--record[/red2] requires [red2]--replay-dir[/red2] to be specified")

        # Set replay directory if replay file is specified
        if self.config.replay_file and not self.config.replay_dir:
            self.config.replay_dir = os.path.dirname(os.path.dirname(self.config.replay_file))

    def parse_hostgroup(self, cfg, section, config_file) -> list[HostGroupMember]:
        hosts = []
        for key in cfg.options(section):
            host_json = cfg.get(section, key).strip()

            # Validate that the host's JSON is valid
            try:
                host_data = json.loads(host_json)
                if not host_data:
                    self.exit(f"{config_file}: Hostgroup [red2]{section}[/red2] has no data for key [red2]{key}[/red2]")
            except json.JSONDecodeError:
                self.exit(
                    f"{config_file}: Invalid JSON value for hostgroup [red2]{section}[/red2], key [red2]{key}[/red2]"
                )

            host = host_data.get("host")
            port = self.config.port
            if ":" in host:
                host, port = host.split(":")

            tab_title = host_data.get("tab_title")
            credential_profile = host_data.get("credential_profile")
            if credential_profile and credential_profile not in self.config.credential_profiles:
                self.exit(
                    f"{config_file}: Credential profile [red2]{credential_profile}[/red2] "
                    f"not found for hostgroup [red2]{section}[/red2], key [red2]{key}[/red2]"
                )

            hosts.append(
                HostGroupMember(tab_title=tab_title, host=host, port=port, credential_profile=credential_profile)
            )

        if not hosts:
            self.exit(
                f"{config_file}: Hostgroup [red2]{section}[/red2] cannot be loaded because "
                f"it doesn't have any hosts listed under its section in Dolphie's config"
            )

        return hosts

    def parse_credential_profile(self, cfg: RawConfigParser, section: str):
        # Options that can be set directly
        credential_profile_options = [
            "host",
            "port",
            "user",
            "password",
            "socket",
            "ssl_mode",
            "ssl_ca",
            "ssl_cert",
            "ssl_key",
        ]

        # All options. mycnf_file and login_path are processed instead of directly set
        supported_options = credential_profile_options + ["mycnf_file", "login_path"]

        credential_name = section.split("credential_profile_")[1]
        credential = CredentialProfile(name=credential_name)

        for key in cfg.options(section):
            value = cfg.get(section, key).strip()

            if key in credential_profile_options:
                setattr(credential, key, value)
                self.add_to_debug_options(f"cred profile setup - {credential_name}", key, value)
            elif key == "mycnf_file":
                if not os.path.isfile(value):
                    self.exit(
                        f"mycnf file [red2]{value}[/red2] for credential profile [red2]{section}[/red2] does not exist"
                    )

                # Parse client options from a my.cnf file
                mycnf = RawConfigParser()
                mycnf.read(value)
                if mycnf.has_section("client"):
                    for option in credential_profile_options:
                        if mycnf.has_option("client", option):
                            option_value = mycnf.get("client", option)

                            setattr(credential, option, option_value)
                            self.add_to_debug_options(
                                f"cred profile setup - {credential_name}", f"{key}/{option}", option_value
                            )
                else:
                    self.exit(
                        f"mycnf file [red2]{value}[/red2] for credential profile [red2]{section}[/red2] "
                        "does not have a client section"
                    )
            elif key == "login_path":
                # Parse login path options from a login path via mysql_config_editor
                try:
                    login_path_data = myloginpath.parse(value)
                    for option in credential_profile_options:
                        if option in login_path_data:
                            option_value = login_path_data[option]

                            setattr(credential, option, option_value)
                            self.add_to_debug_options(
                                f"cred profile setup - {credential_name}", f"{key}/{option}", option_value
                            )
                except Exception as e:
                    self.exit(f"Error reading login path file for credential profile [red2]{section}[/red2]: {e}")
            else:
                self.exit(
                    f"Invalid option [red2]{key}[/red2] found in credential profile [red2]{section}[/red2]. "
                    f"Supported options are: {', '.join(supported_options)}"
                )

        # If section has no options listed
        if not any(getattr(credential, option) for option in credential_profile_options):
            self.exit(
                f"Credential profile [red2]{credential_name}[/red2] has no valid options set. "
                f"Supported options are: {', '.join(supported_options)}"
            )

        credential.ssl = self.create_ssl_object(credential.__dict__)
        if credential.ssl:
            self.add_to_debug_options(f"cred profile setup - {credential_name}", "ssl object", credential.ssl)

        self.config.credential_profiles[credential_name] = credential

    def create_ssl_object(self, data: dict) -> dict:
        ssl_payload = {}

        ssl_mode = data.get("ssl_mode")
        ssl_ca = data.get("ssl_ca")
        ssl_cert = data.get("ssl_cert")
        ssl_key = data.get("ssl_key")

        if ssl_mode:
            ssl_mode = ssl_mode.upper()

            if ssl_mode == "REQUIRED":
                ssl_payload["required"] = True
            elif ssl_mode == "VERIFY_CA":
                if not ssl_ca:
                    self.exit("SSL mode [red2]VERIFY_CA[/red2] requires a CA file (--ssl-ca) to be specified")

                ssl_payload["check_hostname"] = False
                ssl_payload["verify_mode"] = True
            elif ssl_mode == "VERIFY_IDENTITY":
                if not ssl_ca:
                    self.exit("SSL mode [red2]VERIFY_IDENTITY[/red2] requires a CA file (--ssl-ca) to be specified")

                ssl_payload["check_hostname"] = True
                ssl_payload["verify_mode"] = True
            else:
                self.exit(f"Unsupported SSL mode [red2]{ssl_mode}[/red2]")

            if ssl_ca:
                ssl_payload["ca"] = ssl_ca
            if ssl_cert:
                ssl_payload["cert"] = ssl_cert
            if ssl_key:
                ssl_payload["key"] = ssl_key

        return ssl_payload

    def verify_config_value(self, option, value, data_type):
        if data_type is bool:
            if value.lower() == "true":
                return True
            elif value.lower() == "false":
                return False
            else:
                self.exit(
                    f"Error with Dolphie config: [red2]{option}[/red2] is a boolean and must either be true/false"
                )
        elif data_type is int:
            try:
                return int(value)
            except ValueError:
                self.exit(f"Error with Dolphie config: [red2]{option}[/red2] is an integer and must be a number")
        else:
            return value

    def exit(self, message):
        self.console.print(f"[indian_red]{message}[/indian_red]")
        sys.exit()



from __future__ import annotations

import json
import os
import sqlite3
from contextlib import closing
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Any

import orjson
import zstandard as zstd
from dolphie.DataTypes import (
    ConnectionSource,
    ProcesslistThread,
    ProxySQLProcesslistThread,
)
from dolphie.Dolphie import Dolphie
from dolphie.Modules import MetricManager
from dolphie.Modules.Functions import format_bytes, minify_query
from dolphie.Modules.PerformanceSchemaMetrics import PerformanceSchemaMetrics
from loguru import logger


@dataclass
class MySQLReplayData:
    timestamp: str
    system_utilization: dict
    global_status: dict
    global_variables: dict
    binlog_status: dict
    innodb_metrics: dict
    replica_manager: dict
    replication_status: dict
    replication_applier_status: dict
    processlist: dict
    metric_manager: dict
    metadata_locks: dict
    file_io_data: dict
    table_io_waits_data: dict
    statements_summary_data: dict
    group_replication_data: dict
    group_replication_members: dict
    galera_cluster_members: list


@dataclass
class ProxySQLReplayData:
    timestamp: str
    system_utilization: dict
    global_status: dict
    global_variables: dict
    command_stats: dict
    hostgroup_summary: dict
    processlist: dict
    metric_manager: dict


class ReplayManager:
    """ReplayManager class for capturing and replaying Dolphie instance states."""

    # Constants
    PURGE_CHECK_INTERVAL_HOURS = 1
    COMPRESSION_DICT_SIZE = 10 * 1024 * 1024  # 10MB
    COMPRESSION_LEVEL = 5
    COMPRESSION_DICT_SAMPLES = 10

    def __init__(self, dolphie: Dolphie):
        """Initializes the ReplayManager with Dolphie instance and SQLite database settings.

        Args:
            dolphie: The Dolphie instance.
        """
        self.dolphie = dolphie
        # We will increment this to force a new replay file if the schema changes in future versions
        self.schema_version: int = 2
        self.connection: sqlite3.Connection = None
        self.current_replay_id: int = 0  # This is used to keep track of the last primary key read from the database
        self.min_replay_id: int = 0
        self.max_replay_id: int = 0
        self.current_replay_timestamp: str = None  # Only used for dashboard replay section
        self.min_replay_timestamp: str = None
        self.max_replay_timestamp: str = None
        self.total_replay_rows: int = 0
        self.last_purge_time = datetime.now().astimezone() - timedelta(
            hours=self.PURGE_CHECK_INTERVAL_HOURS
        )  # Initialize to an hour ago
        self.replay_file_size: int = 0
        self.dict_samples: list[bytes] = []
        self.global_variable_change_ids: list[int] = []

        self._compression_dict: zstd.ZstdCompressionDict = None
        self._compressor: zstd.ZstdCompressor = None
        self._decompressor: zstd.ZstdDecompressor = None

        # Initialize compressor/decompressor without a dictionary
        self._rebuild_compressor_decompressor()

        # Determine filename used for replay file
        hostname = f"{dolphie.host}_{dolphie.port}"
        if dolphie.replay_file:
            self.replay_file = dolphie.replay_file
        elif dolphie.daemon_mode:
            self.replay_file = f"{dolphie.replay_dir}/{hostname}/daemon.db"
        elif dolphie.record_for_replay:
            timestamp = datetime.now().astimezone().strftime("%Y_%m_%d_%H_%M_%S")
            self.replay_file = f"{dolphie.replay_dir}/{hostname}/{timestamp}.db"
            dolphie.app.notify(
                f"File: [$highlight]{self.replay_file}[/$highlight]",
                title="Recording data",
                timeout=10,
            )
        else:
            # No options specified for replaying, skip initialization
            return

        os.makedirs(os.path.dirname(self.replay_file), mode=0o770, exist_ok=True)
        logger.info(f"Replay SQLite file: {self.replay_file} ({self.dolphie.replay_retention_hours} hours retention)")

        self._initialize_sqlite()
        self._manage_metadata()

    @property
    def compression_dict(self) -> zstd.ZstdCompressionDict | None:
        return self._compression_dict

    @compression_dict.setter
    def compression_dict(self, value: zstd.ZstdCompressionDict | None):
        self._compression_dict = value
        self._rebuild_compressor_decompressor()

    def _rebuild_compressor_decompressor(self):
        """Rebuilds the cached compressor and decompressor with the current compression dictionary."""
        self._compressor = zstd.ZstdCompressor(level=self.COMPRESSION_LEVEL, dict_data=self._compression_dict)
        self._decompressor = zstd.ZstdDecompressor(dict_data=self._compression_dict)

    def _begin_transaction(self) -> None:
        """Begins an immediate transaction for write operations."""
        with closing(self.connection.cursor()) as cursor:
            cursor.execute("BEGIN IMMEDIATE")

    def _commit_transaction(self) -> None:
        """Commits the current transaction."""
        with closing(self.connection.cursor()) as cursor:
            cursor.execute("COMMIT")

    def _rollback_transaction(self) -> None:
        """Rolls back the current transaction."""
        with closing(self.connection.cursor()) as cursor:
            cursor.execute("ROLLBACK")

    def _execute_select_one(self, query: str, params: tuple[Any, ...] = ()) -> tuple[Any, ...] | None:
        """Executes a SELECT query and returns a single row.

        Args:
            query: The SQL SELECT query to execute.
            params: The parameters to bind to the query.

        Returns:
            Optional[Tuple[Any, ...]]: A single row or None if no results.

        Raises:
            sqlite3.Error: If the query execution fails.
        """
        try:
            with closing(self.connection.cursor()) as cursor:
                cursor.execute(query, params)
                return cursor.fetchone()
        except sqlite3.Error as e:
            logger.error(f"Error executing SQLite SELECT query: {e}")
            self.dolphie.app.notify(
                f"Query: {query}\n{e}",
                title="Error executing SQLite query",
                severity="error",
            )
            raise

    def _execute_select_all(self, query: str, params: tuple[Any, ...] = ()) -> list[tuple[Any, ...]]:
        """Executes a SELECT query and returns all rows.

        Args:
            query: The SQL SELECT query to execute.
            params: The parameters to bind to the query.

        Returns:
            List[Tuple[Any, ...]]: A list of rows.

        Raises:
            sqlite3.Error: If the query execution fails.
        """
        try:
            with closing(self.connection.cursor()) as cursor:
                cursor.execute(query, params)
                return cursor.fetchall()
        except sqlite3.Error as e:
            logger.error(f"Error executing SQLite SELECT query: {e}")
            self.dolphie.app.notify(
                f"Query: {query}\n{e}",
                title="Error executing SQLite query",
                severity="error",
            )
            raise

    def _execute_insert(self, query: str, params: tuple[Any, ...] = ()) -> int:
        """Executes an INSERT query and returns the last inserted row ID.

        Args:
            query: The SQL INSERT query to execute.
            params: The parameters to bind to the query.

        Returns:
            int: The last inserted row ID.

        Raises:
            sqlite3.Error: If the query execution fails.
        """
        try:
            with closing(self.connection.cursor()) as cursor:
                cursor.execute(query, params)
                return cursor.lastrowid
        except sqlite3.Error as e:
            logger.error(f"Error executing SQLite INSERT query: {e}")
            self.dolphie.app.notify(
                f"Query: {query}\n{e}",
                title="Error executing SQLite query",
                severity="error",
            )
            raise

    def _execute_modify(self, query: str, params: tuple[Any, ...] = ()) -> int:
        """Executes an UPDATE or DELETE query and returns the number of affected rows.

        Args:
            query: The SQL UPDATE or DELETE query to execute.
            params: The parameters to bind to the query.

        Returns:
            int: The number of affected rows.

        Raises:
            sqlite3.Error: If the query execution fails.
        """
        try:
            with closing(self.connection.cursor()) as cursor:
                cursor.execute(query, params)
                return cursor.rowcount
        except sqlite3.Error as e:
            logger.error(f"Error executing SQLite UPDATE/DELETE query: {e}")
            self.dolphie.app.notify(
                f"Query: {query}\n{e}",
                title="Error executing SQLite query",
                severity="error",
            )
            raise

    def _execute_many(self, query: str, params: list[tuple[Any, ...]]) -> int:
        """Executes a batch of queries and returns the number of affected rows.

        Args:
            query: The SQL query to execute.
            params: A list of parameter tuples for batch execution.

        Returns:
            int: The number of affected rows.

        Raises:
            sqlite3.Error: If the query execution fails.
        """
        try:
            with closing(self.connection.cursor()) as cursor:
                cursor.executemany(query, params)
                return cursor.rowcount
        except sqlite3.Error as e:
            logger.error(f"Error executing SQLite batch query: {e}")
            self.dolphie.app.notify(
                f"Query: {query}\n{e}",
                title="Error executing SQLite query",
                severity="error",
            )
            raise

    def _initialize_sqlite(self):
        """Initializes the SQLite database and creates the necessary tables."""
        database_exists = bool(os.path.exists(self.replay_file))

        self.connection = sqlite3.connect(self.replay_file, isolation_level=None, check_same_thread=False)

        # Lock down the permissions of the replay file
        os.chmod(self.replay_file, 0o660)

        if not database_exists:
            logger.info("Created new SQLite database and connected to it")
        else:
            logger.info("Connected to SQLite")

        # Create replay_data table if it doesn't exist
        self._execute_modify(
            """
            CREATE TABLE IF NOT EXISTS replay_data (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                timestamp DATETIME,
                data BLOB
            )"""
        )
        self._execute_modify("CREATE INDEX IF NOT EXISTS idx_replay_data_timestamp ON replay_data (timestamp)")

        # Create metadata table if it doesn't exist
        self._execute_modify(
            """
            CREATE TABLE IF NOT EXISTS metadata (
                schema_version INTEGER DEFAULT 1,
                host VARCHAR(255),
                port INTEGER,
                host_distro VARCHAR(255),
                connection_source VARCHAR(255),
                dolphie_version VARCHAR(255),
                compression_dict BLOB
            )"""
        )

        # Create variable_changes table if it doesn't exist
        self._execute_modify(
            """
            CREATE TABLE IF NOT EXISTS variable_changes (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                replay_id INTEGER,
                timestamp DATETIME,
                variable_name VARCHAR(255),
                old_value VARCHAR(255),
                new_value VARCHAR(255)
            )"""
        )
        self._execute_modify(
            "CREATE INDEX IF NOT EXISTS idx_variable_changes_timestamp ON variable_changes (timestamp)"
        )
        self._execute_modify(
            "CREATE INDEX IF NOT EXISTS idx_variable_changes_replay_id ON variable_changes (replay_id)"
        )

        # Enable auto-vacuum if it's not already enabled. This will help keep the database file size down.
        result = self._execute_select_one("PRAGMA auto_vacuum")
        if result and result[0] != 1:
            self._execute_modify("PRAGMA auto_vacuum = FULL")
            self._execute_modify("VACUUM")

        self.purge_old_data()

    def purge_old_data(self):
        """Purges data older than the retention period specified by hours_of_retention.
        Only runs if at least an hour has passed since the last purge.
        """
        # Don't purge when not recording, or when loading a replay file (read-only mode)
        if not self.dolphie.record_for_replay or self.dolphie.replay_file:
            return

        current_time = datetime.now().astimezone()
        if (current_time - self.last_purge_time) < timedelta(hours=self.PURGE_CHECK_INTERVAL_HOURS):
            return  # Skip purging if less than an hour has passed

        retention_date = (current_time - timedelta(hours=self.dolphie.replay_retention_hours)).strftime(
            "%Y-%m-%d %H:%M:%S"
        )

        self._execute_modify("DELETE FROM replay_data WHERE timestamp < ?", (retention_date,))
        self._execute_modify("DELETE FROM variable_changes WHERE timestamp < ?", (retention_date,))

        self.last_purge_time = current_time

    def seek_to_previous_id(self) -> bool:
        """Moves current_replay_id back so the next fetch returns the previous row. Gap-safe.

        Returns:
            bool: True if a previous row exists, False if already at the start.
        """
        row = self._execute_select_one(
            "SELECT id, timestamp FROM replay_data WHERE id < ? ORDER BY id DESC LIMIT 1",
            (self.current_replay_id,),
        )
        if row:
            # Set to one before the target so _load_and_parse_replay_data (WHERE id > ?) picks it up
            self.current_replay_id = row[0] - 1
            self.current_replay_timestamp = row[1]
            return True

        return False

    def seek_to_timestamp(self, timestamp: str):
        """Seeks to the specified timestamp in the SQLite database.

        Args:
            timestamp: The timestamp to seek to.
        """
        row = self._execute_select_one(
            "SELECT id, timestamp FROM replay_data WHERE timestamp <= ? ORDER BY timestamp DESC LIMIT 1",
            (timestamp,),
        )
        if not row:
            self.dolphie.app.notify(
                f"No timestamps found on or before [$light_blue]{timestamp}[/$light_blue]",
                severity="error",
                timeout=10,
            )
            return False

        # Set to one before the target so _load_and_parse_replay_data (WHERE id > ?) picks it up
        self.current_replay_id = row[0] - 1
        found_timestamp = row[1]

        if found_timestamp == timestamp:
            self.dolphie.app.notify(
                f"Seeking to timestamp [$light_blue]{timestamp}[/$light_blue]",
                severity="success",
                timeout=10,
            )
        else:
            self.dolphie.app.notify(
                f"Timestamp not found, seeking to closest timestamp [$light_blue]{found_timestamp}[/$light_blue]",
                timeout=10,
            )

        return True

    def _create_new_replay_file(self, new_replay_file: str):
        logger.info(f"Renaming replay file to: {new_replay_file}")

        os.rename(self.replay_file, new_replay_file)

        # Reset compression dict if it's already been set or else the replay file will be corrupted
        self.compression_dict = None

        self._initialize_sqlite()
        self._manage_metadata()

    def _manage_metadata(self):
        """Manages the metadata table with information we care about."""
        # Don't manage metadata when not recording, or when loading a replay file (read-only mode)
        if not self.dolphie.record_for_replay or self.dolphie.replay_file:
            return

        row = self._execute_select_one("SELECT * FROM metadata")
        if row is None:
            self._execute_insert(
                "INSERT INTO metadata (schema_version, host, port, host_distro, connection_source, dolphie_version)"
                "VALUES (?, ?, ?, ?, ?, ?)",
                (
                    self.schema_version,
                    self.dolphie.host,
                    self.dolphie.port,
                    self.dolphie.host_distro,
                    self.dolphie.connection_source,
                    self.dolphie.app_version,
                ),
            )
        else:
            schema_version = row[0]
            if self.dolphie.daemon_mode and schema_version != self.schema_version:
                new_replay_file = f"{self.replay_file}_old_schema_v{schema_version}"
                logger.warning(
                    f"The schema version of the replay file ({schema_version}) differs from this version "
                    f"of Dolphie's schema version ({self.schema_version}). To avoid potential issues, the "
                    f"replay file will be renamed and a new one will be created"
                )

                self._create_new_replay_file(new_replay_file)

                return

            connection_source = row[4]
            if self.dolphie.daemon_mode:
                # Avoid mixing connection sources in the same replay file
                if connection_source != self.dolphie.connection_source:
                    logger.critical(
                        f"The connection source of the daemon's replay file ({connection_source}) "
                        f"differs from the current connection source ({self.dolphie.connection_source}). "
                        "You should never mix connection sources in the same replay file. Please rename "
                        "the daemon's replay file and restart the daemon"
                    )

            host = row[1]
            port = row[2]
            # Add the host's distro to the metadata if it's different than the connection source
            host_distro = f" ({row[3]})" if connection_source != row[3] else ""
            app_version = row[5]
            compress_dict = row[6]

            logger.info(
                f"Replay database metadata - Host: {host}, Port: {port}, Source: {connection_source}{host_distro}, "
                f"Dolphie: {app_version}"
            )

            if compress_dict:
                self.compression_dict = zstd.ZstdCompressionDict(compress_dict)
                logger.info(
                    f"ZSTD compression dictionary loaded (size: {format_bytes(len(compress_dict), color=False)})"
                )

    def verify_replay_file(self):
        """Verifies that the replay file has data to replay and that the schema version matches."""
        if not self.dolphie.replay_file:
            return

        return self._get_replay_file_metadata() and self._verify_replay_has_data()

    def _get_replay_file_metadata(self):
        """Retrieves the replay's metadata from the metadata table.

        Returns:
            bool: True if metadata is found and schema matches; False otherwise.
        """
        row = self._execute_select_one("SELECT * FROM metadata")
        if not row:
            self._notify_error("Metadata not found in replay file", "Error reading replay file")
            return False

        schema_version = row[0]
        if schema_version != self.schema_version:
            self._notify_error(
                f"The schema version of the replay file ({schema_version}) differs from Dolphie's schema version "
                f"({self.schema_version}). Use a compatible version of Dolphie to replay this file",
                "Schema version mismatch",
            )
            return False

        (
            self.dolphie.host,
            self.dolphie.port,
            self.dolphie.host_distro,
            self.dolphie.connection_source,
        ) = row[1:5]
        self.dolphie.host_with_port = f"{self.dolphie.host}:{self.dolphie.port}"

        if row[6]:
            self.compression_dict = zstd.ZstdCompressionDict(row[6])

        return True

    def _verify_replay_has_data(self):
        """Verifies that the replay file has data to replay.

        Returns:
            bool: True if data is found, False if not.
        """
        row = self._execute_select_one("SELECT COUNT(*) FROM replay_data")
        if row and row[0] == 0:
            self._notify_error("File has no data to replay", "No replay data found")
            return False

        return True

    def _notify_error(self, message, title):
        """Helper method to display error notifications."""
        self.dolphie.app.notify(
            f"[b]Replay file[/b]: [$highlight]{self.replay_file}[/$highlight]\n{message}",
            title=title,
            severity="error",
            timeout=10,
        )

    def _train_compression_dict(self) -> bytes:
        """Creates a compression dictionary based on sample data to help with better compression.

        Returns:
            bytes: The created compression dictionary.
        """
        compression_dict = zstd.train_dictionary(
            self.COMPRESSION_DICT_SIZE, self.dict_samples, level=self.COMPRESSION_LEVEL
        )

        logger.info(
            f"ZSTD compression dictionary trained with {len(self.dict_samples)} samples "
            f"(size: {format_bytes(len(compression_dict), color=False)})"
        )

        # Store the compression dictionary in the metadata table to be used with decompression
        self._execute_modify("UPDATE metadata SET compression_dict = ?", (compression_dict.as_bytes(),))

        return compression_dict

    def _condition_metrics(self, metric_manager: MetricManager.MetricManager):
        """Captures the metrics from the metric manager and returns them in a structured format.

        In daemon mode, only the latest value per metric is stored (delta format) to avoid
        serializing the full 10-minute history on every cycle. The replay player detects the
        _delta flag and accumulates values during playback.

        Args:
            metric_manager: The metric manager to capture metrics from.

        Returns:
            dict: A dictionary of captured metrics.
        """
        daemon_mode = self.dolphie.daemon_mode
        connection_source = self.dolphie.connection_source

        if daemon_mode:
            metrics = {
                "datetimes": [metric_manager.datetimes[-1]] if metric_manager.datetimes else [],
                "_delta": True,
            }
        else:
            metrics = {"datetimes": list(metric_manager.datetimes)}

        for (
            metric_instance_name,
            metric_instance_data,
        ) in metric_manager.metrics.__dict__.items():
            if connection_source not in metric_instance_data.connection_source:
                continue
            if connection_source == ConnectionSource.mysql and not metric_instance_data.use_with_replay:
                continue

            metric_entry = metrics.setdefault(metric_instance_name, {})
            for k, v in metric_instance_data.__dict__.items():
                if not hasattr(v, "values") or not v.values:
                    continue

                if daemon_mode:
                    metric_entry[k] = [v.values[-1]]
                else:
                    metric_entry[k] = list(v.values)

        return metrics

    def _prepare_processlist(self) -> list:
        """Prepares the processlist data by extracting thread data and minifying queries.

        Returns:
            list: A list of processlist thread dictionaries with minified queries.
        """
        return [
            {**thread_data, "query": minify_query(thread_data["query"])} if "query" in thread_data else thread_data
            for thread_data in (v.thread_data for v in self.dolphie.processlist_threads.values())
        ]

    def _build_base_data_dict(self, processlist: list) -> dict:
        """Builds the base data dictionary with common data for all connection sources.

        Args:
            processlist: The prepared processlist data.

        Returns:
            dict: The base data dictionary.
        """
        data_dict = {
            "global_status": self.dolphie.global_status,
            "global_variables": self.dolphie.global_variables,
            "processlist": processlist,
            "metric_manager": self._condition_metrics(self.dolphie.metric_manager),
        }

        data_dict["global_status"]["replay_polling_latency"] = self.dolphie.worker_processing_time

        if self.dolphie.system_utilization:
            data_dict["system_utilization"] = self.dolphie.system_utilization

        return data_dict

    def _add_mysql_specific_data(self, data_dict: dict) -> None:
        """Adds MySQL-specific data to the data dictionary.

        Args:
            data_dict: The data dictionary to update.
        """
        # Add the replay_pfs_metrics_last_reset_time to the global status dictionary
        if self.dolphie.pfs_metrics_last_reset_time:
            data_dict["global_status"]["replay_pfs_metrics_last_reset_time"] = (
                datetime.now().astimezone().timestamp() - self.dolphie.pfs_metrics_last_reset_time.timestamp()
            )
        else:
            data_dict["global_status"]["replay_pfs_metrics_last_reset_time"] = 0

        # Add MySQL specific data to the dictionary
        data_dict.update(
            {
                "binlog_status": self.dolphie.binlog_status,
                "innodb_metrics": self.dolphie.innodb_metrics,
                "metadata_locks": self.dolphie.metadata_locks,
            }
        )

        if self.dolphie.replication_status:
            data_dict["replication_status"] = self.dolphie.replication_status

        if self.dolphie.replication_applier_status:
            data_dict["replication_applier_status"] = self.dolphie.replication_applier_status

        if self.dolphie.replica_manager.available_replicas:
            data_dict["replica_manager"] = self.dolphie.replica_manager.available_replicas

        if self.dolphie.group_replication or self.dolphie.innodb_cluster:
            data_dict.update(
                {
                    "group_replication_data": self.dolphie.group_replication_data,
                    "group_replication_members": self.dolphie.group_replication_members,
                }
            )

        if self.dolphie.galera_cluster:
            data_dict["galera_cluster_members"] = self.dolphie.galera_cluster_members

        if self.dolphie.file_io_data and self.dolphie.file_io_data.filtered_data:
            data_dict["file_io_data"] = self.dolphie.file_io_data.filtered_data

        if self.dolphie.table_io_waits_data and self.dolphie.table_io_waits_data.filtered_data:
            data_dict["table_io_waits_data"] = self.dolphie.table_io_waits_data.filtered_data

        if self.dolphie.statements_summary_data and self.dolphie.statements_summary_data.filtered_data:
            data_dict["statements_summary_data"] = self.dolphie.statements_summary_data.filtered_data

    def _serialize_data_dict(self, data_dict: dict) -> bytes:
        """Serializes the data dictionary to bytes using orjson or json as fallback.

        Args:
            data_dict: The data dictionary to serialize.

        Returns:
            bytes: The serialized data.
        """
        # For large numbers, we need to use json instead of orjson to serialize the data
        # to avoid exceeding 64-bit integer limit
        # https://github.com/ijl/orjson/issues/301
        try:
            return orjson.dumps(data_dict)
        except TypeError as e:
            if str(e) == "Integer exceeds 64-bit range":
                return json.dumps(data_dict).encode()
            else:
                raise e

    def _handle_compression_training(self, data_dict_bytes: bytes) -> None:
        """Handles compression dictionary training by collecting samples and training when ready.

        Args:
            data_dict_bytes: The serialized data to use as a training sample.
        """
        if not self.compression_dict:
            if len(self.dict_samples) < self.COMPRESSION_DICT_SAMPLES:
                self.dict_samples.append(data_dict_bytes)
            else:
                self.compression_dict = self._train_compression_dict()
                # Remove the samples to save memory
                del self.dict_samples

    def _insert_replay_data(self, timestamp: str, data_dict_bytes: bytes) -> None:
        """Inserts the replay data into the database and handles variable change linkage.

        Args:
            timestamp: The timestamp of the capture.
            data_dict_bytes: The serialized and compressed data to insert.
        """
        try:
            # Begin transaction for atomic insert and update
            self._begin_transaction()

            # Execute the SQL insert using the constructed dictionary
            self.current_replay_id = self._execute_insert(
                "INSERT INTO replay_data (timestamp, data) VALUES (?, ?)",
                (
                    timestamp,
                    self._compressor.compress(data_dict_bytes),
                ),
            )

            # Update the variable_changes table with the data of the replay row so they're linked
            if self.global_variable_change_ids:
                self._execute_many(
                    "UPDATE variable_changes SET replay_id = ?, timestamp = ? WHERE id = ?",
                    [(self.current_replay_id, timestamp, id) for id in self.global_variable_change_ids],
                )

                # Clear the list of global variable change IDs now that they've been linked
                self.global_variable_change_ids = []

            # Commit the transaction
            self._commit_transaction()

        except Exception as e:
            # Rollback on any error
            self._rollback_transaction()
            logger.error(f"Error inserting replay data: {e}")
            raise

        self.purge_old_data()

        if not self.dolphie.daemon_mode:
            self.replay_file_size = os.path.getsize(self.replay_file)

    def capture_state(self):
        """Captures the current state of the Dolphie instance and stores it in the SQLite database."""
        # Don't capture when not recording, or when loading a replay file (read-only mode)
        if not self.dolphie.record_for_replay or self.dolphie.replay_file:
            return

        # Prepare processlist data
        processlist = self._prepare_processlist()
        timestamp = datetime.now().astimezone().strftime("%Y-%m-%d %H:%M:%S")

        # Build base data dictionary
        data_dict = self._build_base_data_dict(processlist)

        # Add connection-source specific data
        if self.dolphie.connection_source == ConnectionSource.mysql:
            self._add_mysql_specific_data(data_dict)
        else:
            data_dict.update(
                {
                    "command_stats": self.dolphie.proxysql_command_stats,
                    "hostgroup_summary": self.dolphie.proxysql_hostgroup_summary,
                }
            )

        # Serialize and compress the data
        data_dict_bytes = self._serialize_data_dict(data_dict)
        self._handle_compression_training(data_dict_bytes)

        # Insert into database
        self._insert_replay_data(timestamp, data_dict_bytes)

    def _update_replay_metadata_cache(self) -> bool:
        """Updates the replay metadata (min/max timestamps and IDs, total rows).

        Uses two O(1) primary key lookups instead of a full table scan.
        total_replay_rows is derived from the ID range since IDs are contiguous
        (rows are only ever purged from the beginning).

        Returns:
            bool: True if metadata was successfully updated, False otherwise.
        """
        min_row = self._execute_select_one("SELECT id, timestamp FROM replay_data ORDER BY id LIMIT 1")
        if not min_row:
            return False

        max_row = self._execute_select_one("SELECT id, timestamp FROM replay_data ORDER BY id DESC LIMIT 1")

        self.min_replay_id = min_row[0]
        self.min_replay_timestamp = min_row[1]
        self.max_replay_id = max_row[0]
        self.max_replay_timestamp = max_row[1]
        self.total_replay_rows = self.max_replay_id - self.min_replay_id + 1

        return True

    def _load_and_parse_replay_data(self) -> tuple[str, dict] | None:
        """Loads the next replay data row from the database and parses it.

        Returns:
            Optional[Tuple[str, dict]]: A tuple of (timestamp, data_dict) or None if no data available.
        """
        # Get the next row
        row = self._execute_select_one(
            "SELECT id, timestamp, data FROM replay_data WHERE id > ? ORDER BY id LIMIT 1",
            (self.current_replay_id,),
        )
        if not row:
            return None

        self.current_replay_id = row[0]
        self.current_replay_timestamp = row[1]

        # Decompress and parse the JSON data
        try:
            data = orjson.loads(self._decompressor.decompress(row[2]))
            return row[1], data
        except Exception as e:
            self.dolphie.app.notify(str(e), title="Error parsing replay data", severity="error")
            return None

    def _build_processlist_from_data(self, processlist_data: list, thread_class) -> dict:
        """Builds a processlist dictionary from raw data using the specified thread class.

        Args:
            processlist_data: List of thread data dictionaries.
            thread_class: The class to use for creating thread objects (ProcesslistThread or ProxySQLProcesslistThread).

        Returns:
            dict: Dictionary mapping thread IDs to thread objects.
        """
        return {str(thread_data["id"]): thread_class(thread_data) for thread_data in processlist_data}

    def _create_mysql_replay_data(self, timestamp: str, data: dict) -> MySQLReplayData:
        """Creates a MySQLReplayData object from parsed replay data.

        Args:
            timestamp: The timestamp of the replay data.
            data: The parsed data dictionary.

        Returns:
            MySQLReplayData: The constructed replay data object.
        """
        processlist = self._build_processlist_from_data(data["processlist"], ProcesslistThread)

        # Create Performance Schema metrics objects
        file_io_data = PerformanceSchemaMetrics({}, "file_io", "FILE_NAME")
        file_io_data.filtered_data = data.get("file_io_data", {})

        table_io_waits = PerformanceSchemaMetrics({}, "table_io", "OBJECT_TABLE")
        table_io_waits.filtered_data = data.get("table_io_waits_data", {})

        statements_summary_data = PerformanceSchemaMetrics({}, "statements_summary", "digest")
        statements_summary_data.filtered_data = data.get("statements_summary_data", {})

        return MySQLReplayData(
            timestamp=timestamp,
            system_utilization=data.get("system_utilization", {}),
            global_status=data.get("global_status", {}),
            global_variables=data.get("global_variables", {}),
            metric_manager=data.get("metric_manager", {}),
            binlog_status=data.get("binlog_status", {}),
            innodb_metrics=data.get("innodb_metrics", {}),
            replica_manager=data.get("replica_manager", {}),
            replication_status=data.get("replication_status", {}),
            replication_applier_status=data.get("replication_applier_status", {}),
            metadata_locks=data.get("metadata_locks", {}),
            processlist=processlist,
            group_replication_data=data.get("group_replication_data", {}),
            group_replication_members=data.get("group_replication_members", {}),
            galera_cluster_members=data.get("galera_cluster_members", []),
            file_io_data=file_io_data,
            table_io_waits_data=table_io_waits,
            statements_summary_data=statements_summary_data,
        )

    def _create_proxysql_replay_data(self, timestamp: str, data: dict) -> ProxySQLReplayData:
        """Creates a ProxySQLReplayData object from parsed replay data.

        Args:
            timestamp: The timestamp of the replay data.
            data: The parsed data dictionary.

        Returns:
            ProxySQLReplayData: The constructed replay data object.
        """
        processlist = self._build_processlist_from_data(data["processlist"], ProxySQLProcesslistThread)

        return ProxySQLReplayData(
            timestamp=timestamp,
            system_utilization=data.get("system_utilization", {}),
            global_status=data.get("global_status", {}),
            global_variables=data.get("global_variables", {}),
            metric_manager=data.get("metric_manager", {}),
            command_stats=data.get("command_stats", {}),
            hostgroup_summary=data.get("hostgroup_summary", {}),
            processlist=processlist,
        )

    def get_next_refresh_interval(
        self,
    ) -> MySQLReplayData | ProxySQLReplayData | None:
        """Gets the next refresh interval's data from the SQLite database and returns it as a ReplayData object.

        Returns:
            ReplayData: The next replay data.
        """
        # Always update metadata cache when replaying to account for new data being recorded
        # This ensures max_replay_timestamp and max_replay_id reflect the latest state
        if not self._update_replay_metadata_cache():
            return None

        # Load and parse the next replay data
        result = self._load_and_parse_replay_data()
        if not result:
            return None

        timestamp, data = result

        # Create and return the appropriate replay data object based on connection source
        if self.dolphie.connection_source == ConnectionSource.mysql:
            return self._create_mysql_replay_data(timestamp, data)
        elif self.dolphie.connection_source == ConnectionSource.proxysql:
            return self._create_proxysql_replay_data(timestamp, data)
        else:
            self.dolphie.app.notify("Invalid connection source for replay data", severity="error")
            return None

    def fetch_delta_metrics_for_window(
        self, target_id: int, window_minutes: int = MetricManager.MetricManager.ROLLING_WINDOW_MINUTES
    ) -> list[dict]:
        """Fetches metric_manager deltas for the time window ending at the target replay ID.

        This is used during replay seek/backward navigation in delta mode to rebuild
        the 10-minute rolling metric window up to the target position.

        Args:
            target_id: The replay ID to build the window up to.
            window_minutes: The size of the rolling window in minutes.

        Returns:
            list[dict]: A list of metric_manager dicts from oldest to newest within the window.
        """
        if not self.current_replay_timestamp:
            return []

        # Calculate the window start timestamp
        target_dt = datetime.fromisoformat(self.current_replay_timestamp)
        window_start = (target_dt - timedelta(minutes=window_minutes)).strftime("%Y-%m-%d %H:%M:%S")

        # Fetch all rows in the window up to and including the target
        rows = self._execute_select_all(
            "SELECT data FROM replay_data WHERE timestamp >= ? AND id <= ? ORDER BY id",
            (window_start, target_id),
        )

        # Extract just the metric_manager data from each row
        metrics_list = []
        for row in rows:
            try:
                data = orjson.loads(self._decompressor.decompress(row[0]))
                metric_manager = data.get("metric_manager", {})
                if metric_manager:
                    metrics_list.append(metric_manager)
            except Exception:
                continue

        return metrics_list

    def fetch_global_variable_changes_for_current_replay_id(self):
        """Fetches global variable changes for the current replay ID."""
        rows = self._execute_select_all(
            "SELECT timestamp, variable_name, old_value, new_value FROM variable_changes WHERE replay_id = ?",
            (self.current_replay_id,),
        )

        for timestamp, variable, old_value, new_value in rows:
            # read_only notification is handled by monitor_read_only_change() in app.py
            if variable == "read_only":
                continue

            self.dolphie.app.notify(
                f"[b][$dark_yellow]{variable}[/b][/$dark_yellow]\n"
                f"Timestamp: [$light_blue]{timestamp}[/$light_blue]\n"
                f"Old Value: [$highlight]{old_value}[/$highlight]\n"
                f"New Value: [$highlight]{new_value}[/$highlight]",
                title="Global Variable Change",
                severity="warning",
                timeout=10,
            )

    def fetch_all_global_variable_changes(self) -> list:
        """Fetches all global variable changes for command 'V'."""
        rows = self._execute_select_all(
            "SELECT timestamp, variable_name, old_value, new_value FROM variable_changes ORDER BY timestamp"
        )

        return rows

    def capture_global_variable_change(self, variable_name: str, old_value: str, new_value: str):
        """Captures a global variable change and stores it in the SQLite database.

        Args:
            variable_name: The name of the variable that changed.
            old_value: The old value of the variable.
            new_value: The new value of the variable.
        """
        # Don't capture when not recording, or when loading a replay file (read-only mode)
        if not self.dolphie.record_for_replay or self.dolphie.replay_file:
            return

        timestamp = datetime.now().astimezone().strftime("%Y-%m-%d %H:%M:%S")

        last_row_id = self._execute_insert(
            "INSERT INTO variable_changes (timestamp, variable_name, old_value, new_value) VALUES (?, ?, ?, ?)",
            (timestamp, variable_name, old_value, new_value),
        )

        # Keep track of the primary key of the global variable change so we can link it to the replay data
        self.global_variable_change_ids.append(last_row_id)



from dolphie.Modules.Functions import format_bytes, format_number
from dolphie.Modules.TabManager import Tab
from textual.widgets import DataTable


def create_panel(tab: Tab) -> DataTable:
    dolphie = tab.dolphie

    columns = {
        "hostgroup": {"name": "Hostgroup", "width": None, "format": None},
        "srv_host": {"name": "Backend Host", "width": 35, "format": None},
        "srv_port": {"name": "Port", "width": 5, "format": None},
        "status": {"name": "Status", "width": None, "format": None},
        "weight": {"name": "Weight", "width": None, "format": None},
        "use_ssl": {"name": "SSL", "width": 5, "format": None},
        "ConnUsed": {"name": "Conn Used", "width": 11, "format": "number"},
        "ConnFree": {"name": "Conn Free", "width": 10, "format": "number"},
        "ConnOK": {"name": "Conn OK", "width": 10, "format": "number"},
        "ConnERR": {"name": "Conn ERR", "width": 10, "format": "number"},
        "MaxConnUsed": {"name": "Max Conn", "width": 11, "format": "number"},
        "Queries_per_sec": {"name": "Queries/s", "width": 10, "format": "number"},
        "Bytes_data_sent_per_sec": {
            "name": "Data Sent/s",
            "width": 11,
            "format": "bytes",
        },
        "Bytes_data_recv_per_sec": {
            "name": "Data Recvd/s",
            "width": 12,
            "format": "bytes",
        },
        "Latency_us": {"name": "Latency (ms)", "width": 12, "format": "time"},
    }

    # Refresh optimization
    hostgroup_summary_datatable = tab.proxysql_hostgroup_summary_datatable

    column_keys = []
    column_names = []
    column_formats = []

    # Add columns to the datatable if it is empty
    if not hostgroup_summary_datatable.columns:
        for column_key, column_data in columns.items():
            hostgroup_summary_datatable.add_column(column_data["name"], key=column_key, width=column_data["width"])

    for column_key, column_data in columns.items():
        column_keys.append(column_key)
        column_names.append(column_data["name"])
        column_formats.append(column_data["format"])

    changed = False

    with dolphie.app.batch_update():
        # Remove stale rows first
        if dolphie.proxysql_hostgroup_summary:
            current_rows = {
                f"{row['hostgroup']}_{row['srv_host']}_{row['srv_port']}" for row in dolphie.proxysql_hostgroup_summary
            }
            existing_rows = set(hostgroup_summary_datatable.rows.keys())

            rows_to_remove = existing_rows - current_rows
            if rows_to_remove:
                changed = True
                if len(rows_to_remove) > len(dolphie.proxysql_hostgroup_summary):
                    hostgroup_summary_datatable.clear()
                else:
                    for row_id in rows_to_remove:
                        hostgroup_summary_datatable.remove_row(row_id)
        else:
            if hostgroup_summary_datatable.row_count:
                changed = True
                hostgroup_summary_datatable.clear()

        # Iterate through hostgroup summary data
        for row in dolphie.proxysql_hostgroup_summary:
            row_id = f"{row['hostgroup']}_{row['srv_host']}_{row['srv_port']}"

            row_values = []
            for column_key, column_format in zip(column_keys, column_formats):
                column_value = row.get(column_key, 0)

                if column_format == "time":
                    column_value = f"{round(int(column_value) / 1000, 2)}"
                elif column_format == "bytes":
                    column_value = format_bytes(column_value)
                elif column_format == "number":
                    column_value = format_number(column_value)
                elif column_key == "hostgroup":
                    column_value = int(column_value)
                elif column_key == "srv_host":
                    column_value = dolphie.get_hostname(column_value)
                elif column_key == "status":
                    column_value = "[green]ONLINE" if column_value == "ONLINE" else f"[red]{column_value}"
                elif column_key == "use_ssl":
                    column_value = "ON" if column_value == "1" else "OFF"

                if column_key != "hostgroup" and (column_value == "0" or column_value == 0):
                    column_value = "[dark_gray]0"

                row_values.append(column_value)

            if row_id in hostgroup_summary_datatable.rows:
                datatable_row = hostgroup_summary_datatable.get_row(row_id)

                for column_id, column_key in enumerate(column_keys):
                    if row_values[column_id] != datatable_row[column_id]:
                        changed = True
                        hostgroup_summary_datatable.update_cell(row_id, column_key, row_values[column_id])
            else:
                changed = True
                hostgroup_summary_datatable.add_row(*row_values, key=row_id)

        if changed:
            hostgroup_summary_datatable.sort("hostgroup")

    tab.proxysql_hostgroup_summary_title.update(
        f"{dolphie.panels.proxysql_hostgroup_summary.title} "
        f"([$highlight]{hostgroup_summary_datatable.row_count}[/$highlight])"
    )



from dolphie.Modules.Functions import format_bytes, format_time
from dolphie.Modules.TabManager import Tab
from textual.widgets import DataTable


def create_panel(tab: Tab) -> DataTable:
    dolphie = tab.dolphie

    columns = {
        "processlist_id": {"name": "Thread ID", "width": 11, "format": None},
        "percentage_completed": {"name": "Completed", "width": 9, "format": None},
        "memory": {"name": "Memory", "width": 10, "format": "bytes"},
        "started_ago": {"name": "Current Time", "width": 12, "format": "time"},
        "estimated_remaining_time": {
            "name": "Remaining Time",
            "width": 14,
            "format": "time",
        },
        "state": {"name": "State", "width": None, "format": None},
    }

    ddl_datatable = tab.ddl_datatable

    with dolphie.app.batch_update():
        ddl_datatable.clear(columns=True)

        column_keys = []
        column_formats = []
        for column_key, column_data in columns.items():
            ddl_datatable.add_column(column_data["name"], key=column_key, width=column_data["width"])
            column_keys.append(column_key)
            column_formats.append(column_data["format"])

        for ddl in dolphie.ddl:
            row_values = []

            for column_key, column_format in zip(column_keys, column_formats):
                if column_format == "time":
                    value = format_time(ddl[column_key], picoseconds=True)
                elif column_format == "bytes":
                    value = format_bytes(ddl[column_key])
                else:
                    value = ddl[column_key]

                row_values.append(value)

            ddl_datatable.add_row(*row_values, key=ddl["processlist_id"])

    tab.ddl_title.update(f"{dolphie.panels.ddl.title} " f"([$highlight]{ddl_datatable.row_count}[/$highlight])")
